@@ -9,6 +9,7 @@ import json
 import signal
 import sqlite3
 import sys
+import traceback
 from contextlib import asynccontextmanager
 from functools import reduce
 from logging import getLogger
@@ -21,9 +22,13 @@ from gunicorn.arbiter import Arbiter
 from jinja2 import Template
 from tortoise.exceptions import DoesNotExist
 
-from marv.db import DBError, DBNotInitialized, DBVersionError, USERGROUP_REGEX
-from marv.site import UnknownNode, load_sitepackages, make_config
-from marv.utils import echo, err, find_obj, within_pyinstaller_bundle
+import marv_node.run
+from marv.config import ConfigError
+from marv.db import USERGROUP_REGEX, DBError, DBNotInitialized, DBPermissionError, DBVersionError
+from marv.site import SiteError, load_sitepackages, make_config
+from marv.utils import within_pyinstaller_bundle
+from marv_api import ReaderError
+from marv_api.utils import echo, err, find_obj
 from marv_cli import PDB
 from marv_cli import marv as marvcli
 from marv_node.stream import RequestedMessageTooOld
@@ -74,6 +79,8 @@ async def create_site(init=None):
         err(f'{exc!r}\n\n'
             'Existing database is not compatible with this version of MARV. '
             'Check the migration instructions.', exit=1)
+    except (ConfigError, SiteError) as exc:
+        err(f'ERROR: {exc}', exit=1)
     try:
         yield site
     finally:
@@ -194,17 +201,30 @@ def marvcli_serve(host, port, certfile, keyfile, approot):
     config = get_site_config()
 
     async def app_factory():
+        """App factory used inside of worker.
+
+        Note:
+            Terminates execution via sys.exit(4) to prevent master from restarting worker.
+
+        """
         try:
             site = await Site.create(config)
-            application = App(site, app_root=approot).aioapp
-        except (sqlite3.OperationalError, DBNotInitialized) as e:
-            err(f'{e!r}\nDid you run marv init?', exit=1)
-        except DBVersionError as e:
-            err(f'{e!r}\n'
+            try:
+                application = App(site, app_root=approot).aioapp
+            except Exception:  # pylint: disable=broad-except
+                await site.destroy()
+                raise
+        except (sqlite3.OperationalError, DBNotInitialized) as exc:
+            err(f'{exc!r}\nDid you run marv init?', exit=4)
+        except DBVersionError as exc:
+            err(f'{exc!r}\n'
                 'Existing database is not compatible with this version of MARV. '
-                'Check the migration instructions.', exit=1)
-        except PermissionError as e:
-            err(e, exit=1)
+                'Check the migration instructions.', exit=4)
+        except (PermissionError, SiteError) as exc:
+            err(f'ERROR: {exc}', exit=4)
+        except Exception:  # pylint: disable=broad-except
+            traceback.print_exc()
+            sys.exit(4)
         return application
 
     GunicornApplication(app_factory, f'{host}:{port}', certfile, keyfile).run()
@@ -260,16 +280,27 @@ async def marvcli_undiscard(datasets):
 
 
 @marvcli.command('dump')
-@click.argument('dump_file', type=click.File('w'))
+@click.argument('dump_file', type=Path)
 @click_async
 async def marvcli_dump(dump_file):
     """Dump database to json file.
 
     Use '-' for stdout.
     """
+    if str(dump_file) != '-' and dump_file.exists():
+        err('ERROR: Dump file must not exist already!', exit=1)
+
     siteconf = make_config(get_site_config())
-    dump = await Site.Database.dump_database(siteconf.marv.dburi)
-    json.dump(dump, dump_file, sort_keys=True, indent=2)
+    try:
+        dump = await Site.Database.dump_database(siteconf.marv.dburi)
+    except (DBNotInitialized, DBVersionError) as exc:
+        err(f'ERROR: {exc}', exit=1)
+
+    if str(dump_file) == '-':
+        json.dump(dump, sys.stdout, sort_keys=True, indent=2)
+    else:
+        with dump_file.open('w') as f:
+            json.dump(dump, f, sort_keys=True, indent=2)
 
 
 @marvcli.command('restore')
@@ -363,7 +394,7 @@ async def marvcli_query(ctx, list_tags, collections, discarded,
             click.echo(sep.join(str(x) for x in setids), nl=not null)
 
 
-@marvcli.command('run', short_help='Run nodes for DATASETS')  # noqa: C901
+@marvcli.command('run', short_help='Run nodes for DATASETS')
 @click.option(
     '--node', 'selected_nodes', multiple=True,
     help=(
@@ -498,6 +529,14 @@ async def marvcli_run(  # noqa: C901
         else:
             setids = await site.db.get_datasets_for_collections(collections)
 
+        if not PDB:
+            # TODO: Move signal handling into runner
+            def handle_abort(_1, _2):
+                marv_node.run.setabort()
+
+            signal.signal(signal.SIGINT, handle_abort)
+            signal.signal(signal.SIGTERM, handle_abort)
+
         for setid in setids:
             if PDB:
                 await site.run(setid, selected_nodes, deps, force, keep,
@@ -508,8 +547,8 @@ async def marvcli_run(  # noqa: C901
                     await site.run(setid, selected_nodes, deps, force, keep,
                                    force_dependent, update_detail, update_listing,
                                    excluded_nodes, cachesize=cachesize)
-                except UnknownNode as e:
-                    ctx.fail(f'Collection {e.args[0]} has no node {e.args[1]}')
+                except ConfigError as exc:
+                    err(f'ERROR: {exc}', exit=1)
                 except DoesNotExist:
                     click.echo(f'ERROR: unknown {setid!r}', err=True)
                     if not keep_going:
@@ -522,12 +561,13 @@ async def marvcli_run(  # noqa: C901
     See https://ternaris.com/marv-robotics/docs/patterns.html#reduce-separately
     """, err=True)
                     ctx.abort()
-                except BaseException as e:  # pylint: disable=broad-except
+                except marv_node.run.Aborted:
+                    ctx.abort()
+                except ReaderError as e:
+                    errors.append(setid)
+                    log.error('Reader error for dataset %s: %s', setid, e)
+                except Exception as e:  # pylint: disable=broad-except
                     errors.append(setid)
-                    if isinstance(e, KeyboardInterrupt):
-                        log.warning('KeyboardInterrupt: aborting')
-                        raise
-
                     if isinstance(e, DirectoryAlreadyExists):
                         click.echo(f"""
     ERROR: Directory for node run already exists:
@@ -552,23 +592,38 @@ async def marvcli_run(  # noqa: C901
 async def marvcli_scan(dry_run):
     """Scan for new and changed files."""
     async with create_site() as site:
-        await site.scan(dry_run)
+        try:
+            await site.scan(dry_run)
+        except ConfigError as exc:
+            err(f'ERROR: {exc}', exit=1)
 
 
 @marvcli.command('tag')
 @click.option('--add', multiple=True, help='Tags to add')
 @click.option('--rm', '--remove', 'remove', multiple=True, help='Tags to remove')
+@click.option(
+    '--strict',
+    is_flag=True,
+    help=('By default tagging via CLI is idempotent; in strict mode it will fail if tags'
+          ' are repeatedly added or removed.'),
+)
 @click.argument('datasets', nargs=-1)
 @click.pass_context
 @click_async
-async def marvcli_tag(ctx, add, remove, datasets):
+async def marvcli_tag(ctx, add, remove, strict, datasets):
     """Add or remove tags to datasets."""
     if not any([add, remove]) or not datasets:
         click.echo(ctx.get_help())
         ctx.exit(1)
 
     async with create_site() as site:
-        await site.db.update_tags_by_setids(datasets, add, remove)
+        try:
+            await site.db.update_tags_by_setids(datasets, add, remove, idempotent=not strict)
+        except DBPermissionError:
+            if strict:
+                err('ERROR: --strict prevented add of existing or remove of non-existing tag(s).',
+                    exit=1)
+            raise
 
 
 @marvcli.group('comment')
@@ -635,7 +690,7 @@ async def marvcli_show(datasets):
       marv query --col=* | xargs marv show   # show all datasets
     """
     async with create_site() as site:
-        datasets = await site.db.get_datasets_by_setids(datasets, prefetch=['files'])
+        datasets = await site.db.get_datasets_by_setids(datasets, prefetch=['files'], user='::')
         yamldoc = SHOW_TEMPLATE.render(datasets=datasets)
         echo(yamldoc, end='')
 
@@ -830,7 +885,7 @@ def marvcli_pip_install(pipargs):
     load_sitepackages(sitepackages)
     ensure_python(siteconf, sitepackages)
     from pip._internal.main import main as pip_main  # pylint: disable=import-outside-toplevel
-    sys.argv = [sys.executable, 'install', '--prefix', config.marv.venv, *pipargs]
+    sys.argv = [sys.executable, 'install', '--prefix', str(config.marv.venv), *pipargs]
     sys.exit(pip_main())
 
 
