@@ -17,15 +17,14 @@ from marv import utils
 from marv.config import ConfigError, calltree, getdeps, make_funcs, parse_function
 from marv.db import scoped_session
 from marv.model import Collection as CollectionModel
-from marv.model import Comment, Dataset, File
-from marv.model import make_listing_model, make_table_descriptors
+from marv.model import Comment, Dataset, File, make_listing_model, make_table_descriptors
+from marv_api.setid import SetID
+from marv_api.utils import find_obj
 from marv_detail import FORMATTER_MAP, detail_to_dict
 from marv_detail.types_capnp import Detail  # pylint: disable=no-name-in-module
 from marv_node.node import Node
-from marv_node.setid import SetID
 from marv_store import Store
 
-
 FILTER_OPERATORS = \
     'lt le eq ne ge gt substring startswith any all substring_any words'.split()
 FILTER_MAP = {
@@ -50,7 +49,7 @@ def make_filter_spec(line):
     fields = [x.strip() for x in line.split('|', 4)]
     name, title, ops, value_type, function = fields
     ops = ops.split()
-    return FilterSpec(name, title, ops, value_type, function)
+    return FilterSpec(f'f_{name}', title, ops, value_type, function)
 
 
 def make_listing_column(line):
@@ -71,25 +70,21 @@ def make_summary_item(line):
     return SummaryItem(name, title, formatter, islist, function)
 
 
-def flatten_syntax(functree):
-    functree = (flatten_syntax(x) if isinstance(x, tuple) else x for x in functree)
-    functree = (x if len(x) > 1 else x[0] for x in functree)
-    functree = ({} if x[1] == [] else
-                ('map', {}, ('idx', x[1][0]), x[1][1]) if x[0] == 'rows' else
-                x for x in functree)
-    return tuple(functree)
+def postprocess_functree(name, arguments):
+    if name == 'rows':
+        if len(arguments) == 0:
+            return {}
 
+        try:
+            field, default = arguments
+        except ValueError:
+            raise ConfigError(
+                "'rows' takes zero or two arguments, column name and default value",
+            ) from None
 
-def parse_summary(items):
-    summary = []
-    for item in items:
-        functree, pos = parse_function(item.function)
-        assert pos == len(item.function)
-        dct = item._asdict()
-        functree = flatten_syntax(functree)
-        dct['function'] = functree
-        summary.append(dct)
-    return summary
+        return ('map', {}, ('idx', field), default)
+
+    return (name, *(postprocess_functree(*x) for x in arguments))
 
 
 def rowdumps(*args, **kw):
@@ -164,8 +159,7 @@ class Collection:
 
     @property
     def compare(self):
-        # TODO: should we load?
-        return self.section.compare
+        return find_obj(self.section.compare) if self.section.compare else None
 
     @cached_property
     def detail_deps(self):
@@ -179,12 +173,19 @@ class Collection:
     @cached_property
     def detail_sections(self):
         nodes = self.nodes
-        return [nodes[x] for x in self.section.detail_sections]
+        try:
+            return [nodes[x] for x in self.section.detail_sections]
+        except KeyError as exc:
+            raise ConfigError(f'Collection {self.name!r}: unknown node {exc} in detail_sections')
 
     @cached_property
     def detail_summary_widgets(self):
         nodes = self.nodes
-        return [nodes[x] for x in self.section.detail_summary_widgets]
+        try:
+            return [nodes[x] for x in self.section.detail_summary_widgets]
+        except KeyError as exc:
+            raise ConfigError(
+                f'Collection {self.name!r}: unknown node {exc} in detail_summary_widgets')
 
     @cached_property
     def detail_title(self):
@@ -194,10 +195,9 @@ class Collection:
     def filter_functions(self):
         funcs = []
         for spec in self.filter_specs.values():
-            if spec.name in ('comments', 'status', 'tags'):
+            if spec.name in ('f_comments', 'f_status', 'f_tags'):
                 continue
-            functree, pos = parse_function(spec.function)
-            assert pos == len(spec.function)
+            functree = parse_function(spec.function)
             funcs.append((spec, functree))
         return funcs
 
@@ -206,8 +206,8 @@ class Collection:
         specs = [make_filter_spec(line) for line in self.section.filters]
         names = [x.name for x in specs]
         assert all(re.match('^[a-z0-9_]+$', x) for x in names)
-        assert 'setid' in names
-        assert 'tags' in names
+        assert 'f_setid' in names
+        assert 'f_tags' in names
         return OrderedDict((x.name, x) for x in specs)
 
     @cached_property
@@ -225,8 +225,7 @@ class Collection:
     def listing_functions(self):
         funcs = []
         for col in self.listing_columns:
-            functree, pos = parse_function(col.function)
-            assert pos == len(col.function)
+            functree = parse_function(col.function)
             funcs.append((col, functree))
         return funcs
 
@@ -244,27 +243,26 @@ class Collection:
         linemap = {}
         for line in self.section.nodes:
             try:
-                nodename, node = utils.find_obj(line, True)
-            except AttributeError:
-                raise ConfigError(self.section, 'nodes', 'Cannot find node %s' % line)
+                nodename, node = find_obj(line, True)
+            except (AttributeError, ModuleNotFoundError):
+                raise ConfigError(f'Collection {self.name!r} cannot find node {line!r}')
 
             node = Node.from_dag_node(node)
 
             if node in linemap:
-                raise ConfigError(self.section, 'nodes',
-                                  '%s already listed as %s' %
-                                  (line, linemap[node]))
+                raise ConfigError(
+                    f'Collection {self.name!r} node {line!r} already listed as {linemap[node]!r}')
             linemap[node] = line
             if nodename in nodes:
-                raise ConfigError(self.section, 'nodes', 'duplicate name %s' % (nodename,))
+                raise ConfigError(f'Collection {self.name!r} duplicate name {nodename!r}')
             if not node.schema:
-                raise ConfigError(self.section, 'nodes', '%s does not define a schema!' % (line,))
+                raise ConfigError(f'Collection {self.name!r} node {line!r} does not define schema')
             nodes[nodename] = node
         return nodes
 
     @property
     def scanner(self):
-        return utils.find_obj(self.section.scanner)
+        return find_obj(self.section.scanner)
 
     @property
     def scanroots(self):
@@ -290,17 +288,26 @@ class Collection:
         except IndexError:
             return 'ascending'
         except KeyError:
-            raise ConfigError(self.section, 'listing_sort', f'{sortorder} not in {orders}')
+            raise ConfigError(f'Collection {self.name!r} {sortorder} not in {orders}')
 
     @cached_property
     def summary_items(self):
-        items = [make_summary_item(line)
-                 for line in self.section.listing_summary]
-        return parse_summary(items)
+        summary = []
+        for line in self.section.listing_summary:
+            item = make_summary_item(line)
+            dct = item._asdict()
+            functree = parse_function(item.function)
+            try:
+                functree = postprocess_functree(*functree)
+            except ConfigError as exc:
+                raise ConfigError(f'Collection {self.name!r} listing_summary: {exc}') from None
+            dct['function'] = functree
+            summary.append(dct)
+        return summary
 
     @property
     def section(self):
-        return self.config[f'collection {self.name}']
+        return self.config.collections[self.name]
 
     def __init__(self, config, name, site):
         self.config = config
@@ -362,9 +369,10 @@ class Collection:
             # Scan for new files
             batch = []
             for directory, subdirs, filenames in utils.walk(scanpath):
+                directory = str(directory)  # TODO: for now we don't pass Path into scanner
                 # Ignore directories containing a .marvignore file
                 if os.path.exists(os.path.join(directory, '.marvignore')):
-                    subdirs[:] = []
+                    subdirs.clear()
                     continue
 
                 # Ignore hidden directories and traverse subdirs alphabetically
@@ -372,8 +380,10 @@ class Collection:
 
                 # Ignore hidden and known files
                 known = known_filenames[directory]
-                filenames = {x for x in filenames if x[0] != '.'}
-                filenames = sorted(filenames - known)
+                filenames = sorted(x for x in filenames if x[0] != '.' and x not in known)
+
+                if not filenames and not subdirs:
+                    continue
 
                 for name, files in self.scanner(directory, subdirs, filenames):
                     files = [x if os.path.isabs(x) else os.path.join(directory, x)
@@ -384,8 +394,9 @@ class Collection:
                     else:
                         dataset = await self.make_dataset(connection, files, name)
                         batch.append(dataset)
-                        if len(batch) > 50:
+                        if len(batch) >= 50:
                             await self._upsert_listing(connection, log, batch)
+                            batch.clear()
 
             if not dry_run and batch:
                 await self._upsert_listing(connection, log, batch)
@@ -422,18 +433,18 @@ class Collection:
                 batch.append(dataset)
                 if len(batch) > 50:
                     await Comment.bulk_create(comments, using_db=connection)
-                    comments[:] = []
                     await self._upsert_listing(connection, log, batch)
                     await self._add_tags(connection, tags)
+                    comments.clear()
+                    batch.clear()
+                    tags.clear()
             await Comment.bulk_create(comments, using_db=connection)
-            comments[:] = []
             await self._upsert_listing(connection, log, batch)
             await self._add_tags(connection, tags)
 
     async def _add_tags(self, connection, data):
         add = [(tag, dataset.id) for dataset, tags in data for tag in tags]
         await self.site.db.bulk_tag(add, [], '::', txn=connection)
-        data[:] = []
 
     async def _upsert_listing(self, txn, log, batch, update=False):
         descs = self.table_descriptors
@@ -460,7 +471,6 @@ class Collection:
 
         for dataset in batch:
             log.info(f'{"updated" if update else "added"} %r', dataset)
-        batch[:] = []
 
     async def make_dataset(self, connection, files, name, time_added=None, discarded=False,
                            setid=None, status=0, timestamp=None, _restore=None):
