@@ -4,35 +4,36 @@
 # pylint: disable=too-many-lines
 
 import asyncio
-import hashlib
 import json
 import re
-import secrets
-import string
+import sqlite3
 import sys
 from collections import namedtuple
 from contextlib import asynccontextmanager
 from contextvars import ContextVar
-from datetime import datetime
+from datetime import datetime, timezone
 from functools import reduce
 from itertools import groupby, product
 from logging import getLogger
+from pathlib import Path
 
 import bcrypt
-from pypika import JoinType, Order, SQLLiteQuery as Query, Table, Tables, Tuple, functions as fn
+from pypika import JoinType, Order
+from pypika import SQLLiteQuery as Query
+from pypika import Table, Tables, Tuple
+from pypika import functions as fn
 from pypika.terms import AggregateFunction, Criterion, EmptyCriterion, ValueWrapper
 from tortoise import Tortoise as _Tortoise
 from tortoise.exceptions import DoesNotExist, IntegrityError
 from tortoise.transactions import current_transaction_map
 
-from marv_node.setid import SetID
+from marv_api.setid import SetID
+
 from . import utils
-from .model import Group, Leaf, User
-from .model import STATUS, STATUS_MISSING, STATUS_OUTDATED
+from .model import STATUS, STATUS_MISSING, STATUS_OUTDATED, Group, User
 from .model import __models__ as MODELS
 from .utils import findfirst
 
-
 log = getLogger(__name__)
 
 
@@ -235,7 +236,7 @@ OPS = {
 }
 
 
-def resolve_filter(table, fltr, models):  # noqa: C901
+def resolve_filter(table, fltr, models):  # noqa: C901  pylint: disable=too-many-branches
     if not isinstance(fltr, dict):
         raise FilterError(f'Expected dict not {fltr!r}')
 
@@ -247,7 +248,6 @@ def resolve_filter(table, fltr, models):  # noqa: C901
 
     value = fltr.get('value')
     name = fltr.get('name')
-    tablemeta = findfirst(lambda x: x._meta.table == table._table_name, models)._meta
 
     if operator == 'and':
         return reduce(lambda x, y: x & resolve_filter(table, y, models), value, EmptyCriterion())
@@ -259,6 +259,12 @@ def resolve_filter(table, fltr, models):  # noqa: C901
     if not isinstance(name, str):
         raise FilterError(f"Missing 'name' in {fltr!r}")
 
+    if name.startswith('dataset.') and table._table_name.startswith('l_'):
+        name = name[8:]
+        tablemeta = findfirst(lambda x: x._meta.table == 'dataset', models)._meta
+    else:
+        tablemeta = findfirst(lambda x: x._meta.table == table._table_name, models)._meta
+
     if '.' in name:
         fieldname, subname = name.split('.', 1)
         if fieldname in tablemeta.backward_fk_fields:
@@ -283,10 +289,10 @@ def resolve_filter(table, fltr, models):  # noqa: C901
                                              .select(getattr(through, field.backward_key))
                                              .distinct())
 
-        raise FilterError(f'Field {fieldname!r} not on model {tablemeta.table}')
+        raise FilterError(f'Field {fieldname!r} not on model {tablemeta.table!r}')
 
     if name not in tablemeta.db_fields:
-        raise FilterError(f'Field {name!r} not on model {tablemeta.table}')
+        raise FilterError(f'Field {name!r} not on model {tablemeta.table!r}')
 
     if operator not in OPS:
         raise FilterError(f'Unknown operator {operator!r}')
@@ -297,7 +303,7 @@ def resolve_filter(table, fltr, models):  # noqa: C901
 
 def cleanup_attrs(items):
     return [
-        {k: x[k] for k in x.keys() if k not in ['acn_id', 'password']}
+        {k: x[k] for k in x.keys() if k not in ['acn_id', 'password', 'row']}
         for x in items
     ]
 
@@ -337,8 +343,7 @@ def modelize(items, relations):
 
 def dt_to_sec(dtime):
     """Return seconds since epoch for datetime object."""
-    sec = (datetime.fromisoformat(dtime) - datetime.utcfromtimestamp(0)).total_seconds()
-    return int(sec)
+    return int(datetime.fromisoformat(dtime).replace(tzinfo=timezone.utc).timestamp())
 
 
 async def get_items_for_query(query, txn):
@@ -402,21 +407,6 @@ async def dump_users_groups(tables, dump, txn):
     assert not groups, groups
 
 
-async def dump_leafs(tables, dump, txn):
-    dump['leafs'] = leafs = []
-    if 'leaf' in tables:
-        leaf_t = tables.pop('leaf')
-        items = await get_items_for_query(Query.from_(leaf_t)
-                                          .select('*')
-                                          .orderby(leaf_t.name),
-                                          txn)
-        for leaf in items:
-            del leaf['id']
-            leaf['time_created'] = dt_to_sec(leaf['time_created'])
-            leaf['time_updated'] = dt_to_sec(leaf['time_updated'])
-            leafs.append(leaf)
-
-
 async def dump_acns(tables, dump, txn):  # pylint: disable=unused-argument
     del tables['acn']
 
@@ -504,12 +494,6 @@ async def restore_users(site, dct, txn):
                 await site.db.group_adduser(grp, user['name'], txn=txn)
 
 
-async def restore_leafs(site, dct, txn):
-    leafs = dct.pop('leafs')
-    for leaf in leafs or []:
-        await site.db.leaf_add(_restore=True, txn=txn, **leaf)
-
-
 async def restore_datasets(site, dct, txn):
     datasets = dct.pop('datasets')
     for key, sets in datasets.items():
@@ -527,11 +511,10 @@ class Tortoise(_Tortoise):
 class Database:
     # pylint: disable=too-many-public-methods
 
-    VERSION = '20.06'
+    VERSION = '20.12'
 
     EXPORT_HANDLERS = (
         ({'group', 'user', 'user_group'}, dump_users_groups),
-        ({'leaf'}, dump_leafs),
         ({'acn'}, dump_acns),
         ({'comment'}, dump_comments),
         ({'file'}, dump_files),
@@ -542,15 +525,15 @@ class Database:
 
     IMPORT_HANDLERS = (
         ({'users'}, restore_users),
-        ({'leafs'}, restore_leafs),
         ({'datasets'}, restore_datasets),
     )
 
     MODELS = MODELS
 
-    def __init__(self):
+    def __init__(self, listing_models):
         self.connections = []
         self.connection_queue = asyncio.Queue()
+        self.listing_models = listing_models
 
     async def initialize_connections(self):
         defcon = Tortoise.get_connection('default')
@@ -589,8 +572,8 @@ class Database:
             return False
         return bcrypt.checkpw(password.encode('utf-8'), user.password.encode('utf-8'))
 
-    @run_in_transaction   # noqa: C901
-    async def bulk_um(self, users_add, users_remove, groups_add, groups_remove,    # noqa: C901
+    @run_in_transaction
+    async def bulk_um(self, users_add, users_remove, groups_add, groups_remove,  # noqa: C901
                       groups_add_users, groups_remove_users, txn=None):
         # pylint: disable=too-many-arguments
         try:
@@ -761,64 +744,6 @@ class Database:
             query = query.prefetch_related('users')
         return await query
 
-    @run_in_transaction
-    async def leaf_add(self, name, access_token=None, refresh_token=None, time_created=None,
-                       time_updated=None, _restore=None, txn=None):
-        # pylint: disable=too-many-arguments
-        if _restore:
-            clear_access_token = None
-            clear_refresh_token = None
-        else:
-            alnum = string.ascii_letters + string.digits
-            clear_access_token = ''.join(secrets.choice(alnum) for i in range(20))
-            clear_refresh_token = ''.join(secrets.choice(alnum) for i in range(20))
-            access_token = hashlib.sha256(clear_access_token.encode()).hexdigest()
-            refresh_token = hashlib.sha256(clear_refresh_token.encode()).hexdigest()
-
-        now = datetime.utcnow()  # noqa: DTZ
-        time_created = datetime.utcfromtimestamp(time_created) if time_created else now
-        time_updated = datetime.utcfromtimestamp(time_updated) if time_updated else now
-        try:
-            leaf = await Leaf.create(name=name, access_token=access_token,
-                                     refresh_token=refresh_token, time_created=time_created,
-                                     time_updated=time_updated, using_db=txn)
-            if _restore:
-                leaf_t = Table('leaf')
-                await txn.exq(Query.update(leaf_t)
-                              .set(leaf_t.time_updated, time_updated)
-                              .where(leaf_t.name == name))
-        except IntegrityError:
-            raise ValueError(f'Recording unit {name} exists already')
-
-        return leaf
-
-    @run_in_transaction
-    async def leaf_regentoken(self, name, txn=None):
-        try:
-            leaf = await Leaf.get(name=name).using_db(txn)
-        except DoesNotExist:
-            raise ValueError(f'Recording unit {name} does not exist')
-
-        alnum = string.ascii_letters + string.digits
-        clear_access_token = ''.join(secrets.choice(alnum) for i in range(20))
-        clear_refresh_token = ''.join(secrets.choice(alnum) for i in range(20))
-        leaf.access_token = hashlib.sha256(clear_access_token.encode('utf8')).hexdigest()
-        leaf.refresh_token = hashlib.sha256(clear_refresh_token.encode('utf8')).hexdigest()
-        leaf.time_updated = int(utils.now())
-        await leaf.save(using_db=txn)
-        leaf.clear_access_token = clear_access_token
-        leaf.clear_refresh_token = clear_refresh_token
-        return leaf
-
-    @run_in_transaction
-    async def get_leafs(self, txn=None):
-        return await Leaf.all().using_db(txn).order_by('name')
-
-    @run_in_transaction
-    async def get_leaf_by_token(self, clear_access_token, txn=None):
-        access_token = hashlib.sha256(clear_access_token.encode()).hexdigest()
-        return await Leaf.filter(access_token=access_token).using_db(txn).first()
-
     @run_in_transaction
     async def get_acl(self, model, id, user, default, txn=None):
         # pylint: disable=unused-argument, too-many-arguments
@@ -1129,14 +1054,23 @@ class Database:
                 raise DBPermissionError
 
     @run_in_transaction
-    async def update_tags_by_setids(self, setids, add, remove, txn=None):
+    async def update_tags_by_setids(self, setids, add, remove, idempotent=False, txn=None):
         setids = [str(x) for x in setids]
-        dataset = Table('dataset')
-        colids = await txn.exq(Query.from_(dataset)
-                               .where(dataset.setid.isin(setids))
-                               .select(dataset.id))
-        add = [(tag, colid['id']) for tag, colid in product(add, colids)]
-        remove = [(tag, colid['id']) for tag, colid in product(remove, colids)]
+        dataset, dataset_tag, tag = Tables('dataset', 'dataset_tag', 'tag')
+        tagmap = await txn.exq(Query.from_(dataset)
+                               .join(dataset_tag, how=JoinType.left_outer)
+                               .on(dataset.id == dataset_tag.dataset_id)
+                               .join(tag, how=JoinType.left_outer)
+                               .on(tag.id == dataset_tag.tag_id)
+                               .select(tag.value, dataset.id)
+                               .where(dataset.setid.isin(setids)))
+        current = {tuple(x) for x in tagmap}
+        ids = [x[1] for x in current]
+        add = set(product(add, ids))
+        remove = set(product(remove, ids))
+        if idempotent:
+            add -= current
+            remove &= current
         await self.bulk_tag(add, remove, user='::', txn=txn)
 
     @run_in_transaction
@@ -1196,8 +1130,8 @@ class Database:
             if {'any', 'all'}.intersection(collection.filter_specs[desc.key].operators)
         }
         all_known.update({
-            'status': list(STATUS),
-            'tags': await self.get_all_known_tags_for_collection(collection.name),
+            'f_status': list(STATUS),
+            'f_tags': await self.get_all_known_tags_for_collection(collection.name),
         })
         return all_known
 
@@ -1218,7 +1152,7 @@ class Database:
                      .orderby(tag.value)
         return [x['value'] for x in await txn.exq(query)]
 
-    @run_in_transaction  # noqa: C901
+    @run_in_transaction
     async def get_filtered_listing(self, descs, filters, user, txn=None):  # noqa: C901
         # pylint: disable=too-many-locals,too-many-branches,too-many-statements,unused-argument
         listing, dataset, dataset_tag, tag = \
@@ -1242,14 +1176,14 @@ class Database:
             if isinstance(value, int):
                 value = min(value, sys.maxsize)
 
-            if name == 'comments':
+            if name == 'f_comments':
                 comment = Table('comment')
                 query = query.where(listing.id.isin(Query.from_(comment)
                                                     .select(comment.dataset_id)
                                                     .where(escaped_contains(comment.text, value))))
                 continue
 
-            if name == 'status':
+            if name == 'f_status':
                 status_ids = list(STATUS.keys())
                 bitmasks = [2**status_ids.index(x) for x in value]
                 bitmask = sum(bitmasks)
@@ -1264,7 +1198,7 @@ class Database:
 
                 continue
 
-            if name == 'tags':
+            if name == 'f_tags':
                 if operator == 'any':
                     query = query.where(listing.id.isin(Query.from_(dataset_tag)
                                                         .join(tag)
@@ -1449,16 +1383,24 @@ class Database:
                                  self.get_resolve_value_query(table, relations, '::', None)))
                       .delete())
 
-    @run_in_transaction  # noqa: C901
+    @run_in_transaction
     async def rpc_query(self, model, filters, attrs, order, limit, offset,  # noqa: C901
                         user, txn=None):
         # pylint: disable=too-many-arguments, too-many-statements, too-many-locals, too-many-branches
         result = {}
-        table = Table(model)
-        try:
-            tablemeta = [x for x in self.MODELS if x._meta.table == model][0]._meta
-        except IndexError:
-            raise ValueError(f'there is no model "{model}"')
+        if model.startswith('collection:'):
+            name = f'l_{model[11:]}'
+            table = Table(name)
+            try:
+                tablemeta = [x for x in self.listing_models if x._meta.table == name][0]._meta
+            except IndexError:
+                raise ValueError(f'there is no collection "{name}"')
+        else:
+            table = Table(model)
+            try:
+                tablemeta = [x for x in self.MODELS if x._meta.table == model][0]._meta
+            except IndexError:
+                raise ValueError(f'there is no model "{model}"')
 
         select = [
             getattr(table, k)
@@ -1472,8 +1414,11 @@ class Database:
             select = [table.star]
         query = Query.from_(table).select(*select)
 
-        for filt in filters:
-            query = query.where(resolve_filter(table, filt, self.MODELS))
+        try:
+            for filt in filters:
+                query = query.where(resolve_filter(table, filt, self.MODELS + self.listing_models))
+        except FilterError as e:
+            raise ValueError(*e.args)
 
         def customize_query(query, model, table, through=None):
             if model in ['collection', 'dataset']:
@@ -1510,8 +1455,15 @@ class Database:
                 if 'id' not in select:
                     select.append('id')
 
-            if fieldname in tablemeta.backward_fk_fields:
-                field = tablemeta.fields_map[fieldname]
+            if model.startswith('collection:') and fieldname.startswith('dataset.'):
+                modelfieldname = fieldname[8:]
+                meta = [x for x in self.MODELS if x._meta.table == 'dataset'][0]._meta
+            else:
+                modelfieldname = fieldname
+                meta = tablemeta
+
+            if modelfieldname in meta.backward_fk_fields:
+                field = meta.fields_map[modelfieldname]
                 tablename = field.model_class._meta.table
                 subtable = Table(tablename)
                 if '*' not in select and field.relation_field not in select:
@@ -1527,10 +1479,14 @@ class Database:
                         for x in relres
                         if x[field.relation_field] == prime['id']
                     ]
-                result.setdefault(tablename, []).extend(cleanup_attrs(relres))
+                if model.startswith('collection:'):
+                    reskey = fieldname
+                else:
+                    reskey = tablename
+                result.setdefault(reskey, []).extend(cleanup_attrs(relres))
 
-            elif fieldname in tablemeta.m2m_fields:
-                field = tablemeta.fields_map[fieldname]
+            elif modelfieldname in meta.m2m_fields:
+                field = meta.fields_map[modelfieldname]
                 through = Table(field.through)
                 tablename = field.model_class._meta.table
                 rel = Table(tablename)
@@ -1547,18 +1503,51 @@ class Database:
                 relres = await txn.exq(query)
                 relids = [x['id'] for x in relres]
                 for prime in qres:
-                    prime[fieldname] = [
+                    prime[modelfieldname] = [
                         x[field.forward_key]
                         for x in refs
                         if x[field.backward_key] == prime['id'] and x[field.forward_key] in relids
                     ]
-                result.setdefault(tablename, []).extend(cleanup_attrs(relres))
+                if model.startswith('collection:'):
+                    reskey = fieldname
+                else:
+                    reskey = tablename
+                result.setdefault(reskey, []).extend(cleanup_attrs(relres))
 
             else:
-                raise ValueError(f'field {fieldname} not on model {model}')
+                raise ValueError(f'Field {modelfieldname!r} not on model {model!r}')
 
         return result
 
+    @classmethod
+    def check_db_version(cls, dburi, missing_ok=False):
+        dbpath = Path(dburi.split('sqlite://', 1)[1])
+        if not dbpath.exists():
+            if not missing_ok:
+                raise DBNotInitialized('There is no marv database.')
+            return
+
+        metadata = Table('metadata')
+        required = cls.VERSION
+        try:
+            conn = sqlite3.connect(f'file:{dbpath}?mode=ro', uri=True)
+            with conn:
+                have = conn.execute(Query.from_(metadata)
+                                    .select(metadata.value)
+                                    .where(metadata.key == 'database_version')
+                                    .get_sql())\
+                           .fetchone()[0]
+                if have != required:
+                    raise DBVersionError(
+                        f'DB version on disk "{have}", but "{required}" required.',
+                    )
+        except (sqlite3.OperationalError, TypeError):
+            raise DBVersionError(
+                'DB version is unknown, metadata table or version entry missing.',
+            )
+        finally:
+            conn.close()
+
     async def restore_database(self, site, dct):
         async with scoped_session(self) as txn:
             for keys, handler in self.IMPORT_HANDLERS:
@@ -1581,6 +1570,7 @@ class Database:
         older databases, not matching the current version of marv, can be
         dumped.
         """
+        cls.check_db_version(dburi)
 
         class T2(Tortoise):
             apps = {}
