@@ -4,138 +4,40 @@
 import fcntl
 import os
 import shutil
-import sqlite3
 import sys
-import sysconfig
 from itertools import count, product
 from logging import getLogger
 from pathlib import Path
 from uuid import uuid4
 
-from pkg_resources import resource_filename
-from pypika import SQLLiteQuery as Query, Table
+from pypika import SQLLiteQuery as Query
+from pypika import Table
 
+from marv.utils import within_pyinstaller_bundle
+from marv_api.utils import find_obj
+from marv_node.node import Node
 from marv_node.run import run_nodes
 from marv_store import Store
-from . import utils
+
 from .collection import Collections
-from .config import Config
-from .db import DBNotInitialized, DBVersionError
-from .db import Database, Tortoise, create_or_ignore, scoped_session
+from .config import Config, ConfigError
+from .db import Database, DBNotInitialized, Tortoise, create_or_ignore, scoped_session
 from .model import Dataset, Group, User
 
-
-DEFAULT_NODES = """
-marv_nodes:dataset
-marv_nodes:meta_table
-marv_nodes:summary_keyval
-"""
-
-DEFAULT_FILTERS = """
-name       | Name       | substring         | string   | (get "dataset.name")
-setid      | Set Id     | startswith        | string   | (get "dataset.id")
-size       | Size       | lt le eq ne ge gt | filesize | (sum (get "dataset.files[:].size"))
-status     | Status     | any all           | subset   | (status)
-tags       | Tags       | any all           | subset   | (tags)
-comments   | Comments   | substring         | string   | (comments)
-files      | File paths | substring_any     | string[] | (get "dataset.files[:].path")
-time_added | Added      | lt le eq ne ge gt | datetime | (get "dataset.time_added")
-"""
-
-DEFAULT_LISTING_COLUMNS = """
-name       | Name   | route    | (detail_route (get "dataset.id") (get "dataset.name"))
-size       | Size   | filesize | (sum (get "dataset.files[:].size"))
-status     | Status | icon[]   | (status)
-tags       | Tags   | pill[]   | (tags)
-time_added | Added  | datetime | (get "dataset.time_added")
-"""
-
-DEFAULT_LISTING_SUMMARY = """
-datasets | datasets | int       | (len (rows))
-size     | size     | filesize  | (sum (rows "size" 0))
-"""
-
-DEFAULT_DETAIL_SUMMARY_WIDGETS = """
-summary_keyval
-meta_table
-"""
-
 log = getLogger(__name__)
 
 
-class UnknownNode(Exception):
+class SiteError(Exception):
     pass
 
 
 def make_config(siteconf):
-    sitedir = Path(siteconf).parent
-    py_ver = sysconfig.get_python_version()
-    defaults = {
-        'marv': {
-            'acl': 'marv_webapi.acls:authenticated',
-            'dburi': 'sqlite://' + str(sitedir / 'db' / 'db.sqlite'),
-            'frontenddir': str(sitedir / 'frontend'),
-            'oauth': '',
-            'reverse_proxy': None,
-            'sessionkey_file': str(sitedir / 'sessionkey'),
-            'staticdir': str(sitedir / resource_filename('marv_ludwig', 'static')),
-            'storedir': str(sitedir / 'store'),
-            'upload_checkpoint_commands': '',
-            'venv': str(sitedir / 'venv'),
-            'sitepackages': str(sitedir / 'venv' / 'lib' / f'python{py_ver}' / 'site-packages'),
-            'window_title': '',
-        },
-        'collection': {
-            'compare': None,
-            'detail_summary_widgets': DEFAULT_DETAIL_SUMMARY_WIDGETS,
-            'detail_sections': '',
-            'detail_title': '(get "dataset.name")',
-            'filters': DEFAULT_FILTERS,
-            'listing_columns': DEFAULT_LISTING_COLUMNS,
-            'listing_sort': '| ascending',
-            'listing_summary': DEFAULT_LISTING_SUMMARY,
-            'nodes': DEFAULT_NODES,
-        },
-    }
-    required = {
-        'marv': [
-            'collections',
-        ],
-        'collection': [
-            'scanner',
-            'scanroots',
-        ],
-    }
-    schema = {
-        'marv': {
-            'acl': 'find_obj',
-            'collections': 'space_separated_list',
-            'oauth': 'lines',
-            'staticdir': 'path',
-            'storedir': 'path',
-            'upload_checkpoint_commands': 'cmd_lines',
-        },
-        'collection': {
-            'compare': 'find_obj',
-            'detail_title': 'function',
-            'detail_summary_widgets': 'lines',
-            'detail_sections': 'lines',
-            'filters': 'lines',
-            'listing_columns': 'lines',
-            'listing_sort': 'pipe_separated_list',
-            'listing_summary': 'lines',
-            'nodes': 'lines',
-            'scanroots': 'path_lines',
-        },
-    }
-    return Config.from_file(siteconf, defaults=defaults,
-                            required=required, schema=schema)
+    return Config.from_file(siteconf)
 
 
 def load_sitepackages(sitepackages):
     import site  # pylint: disable=import-outside-toplevel
     site.USER_SITE = sitepackages
-    sitepackages = Path(sitepackages)
     sitepackages.mkdir(parents=True, exist_ok=True)
     if str(sitepackages) not in sys.path:
         sys.path.append(str(sitepackages))
@@ -156,28 +58,20 @@ class Site:
 
     def __init__(self, siteconf):
         self.config = make_config(siteconf)
-        # TODO: maybe ordereddict for meta, or generate multiple keys
-        self.config.marv.oauth = {
-            line.split(' ', 1)[0]: [field.strip() for field in line.split('|')]
-            for line in self.config.marv.oauth
-        }
         self.collections = Collections(config=self.config, site=self)
-        self.db = self.Database()  # pylint: disable=invalid-name
-        self.createdb = True
+        self.db = self.Database([y for x in self.collections.values() for y in x.model])  # pylint: disable=invalid-name
 
     @classmethod
     async def create(cls, siteconf, init=None):  # noqa: C901
         site = cls(siteconf)
-        if utils.within_pyinstaller_bundle():
+        if within_pyinstaller_bundle():
             load_sitepackages(site.config.marv.sitepackages)
+        site.config.marv.resourcedir.mkdir(exist_ok=True)
 
-        assert site.config.marv.dburi.startswith('sqlite://')
-        dbpath = Path(site.config.marv.dburi.split('://', 1)[1])
-        site.createdb = not dbpath.exists()
-        if not site.createdb:
-            site.check_db_version()
-        elif not init:
-            raise DBNotInitialized('There is no marv database.')
+        site.db.check_db_version(site.config.marv.dburi, missing_ok=init)
+
+        dbpath = Path(site.config.marv.dburi.split('sqlite://', 1)[1])
+        store_db_version = not dbpath.exists()
 
         if init:
             site.init_directory()
@@ -194,7 +88,7 @@ class Site:
             log.verbose('Generated %s', site.config.marv.sessionkey_file)
 
         # Generate all dynamic models
-        models = site.db.MODELS + [y for x in site.collections.values() for y in x.model]
+        models = site.db.MODELS + site.db.listing_models
 
         await Tortoise.init(config={
             'connections': {
@@ -217,7 +111,7 @@ class Site:
 
         try:
             if init:
-                await site.init_database()
+                await site.init_database(store_db_version=store_db_version)
 
             async with scoped_session(site.db) as txn:
                 try:
@@ -256,29 +150,6 @@ class Site:
             if e.errno != 17:
                 raise
 
-    def check_db_version(self):
-        dbpath = Path(self.config.marv.dburi.split('://', 1)[1])
-        metadata = Table('metadata')
-        required = self.db.VERSION
-        try:
-            conn = sqlite3.connect(f'file:{dbpath}?mode=ro', uri=True)
-            with conn:
-                have = conn.execute(Query.from_(metadata)
-                                    .select(metadata.value)
-                                    .where(metadata.key == 'database_version')
-                                    .get_sql())\
-                           .fetchone()[0]
-                if have != required:
-                    raise DBVersionError(
-                        f'DB version on disk "{have}", but "{required}" required.',
-                    )
-        except (sqlite3.OperationalError, TypeError):
-            raise DBVersionError(
-                'DB version is unknown, metadata table or version entry missing.',
-            )
-        finally:
-            conn.close()
-
     async def store_db_version(self, txn):
         metadata = Table('metadata')
         await txn.exq(Query.into(metadata)
@@ -298,7 +169,7 @@ class Site:
         for table in sorted(tables, key=len, reverse=True):
             await txn.execute_query(f'DROP TABLE {table};')
 
-    async def init_database(self):
+    async def init_database(self, store_db_version=False):
         async with scoped_session(self.db) as txn:
             await self.drop_listings(txn)
 
@@ -316,7 +187,7 @@ class Site:
                     using_db=txn,
                 )
 
-            if self.createdb:
+            if store_db_version:
                 await self.store_db_version(txn)
 
             await create_or_ignore('acn', id=1, txn=txn)
@@ -377,12 +248,12 @@ class Site:
             selected_nodes.update(collection.detail_deps)
         persistent = collection.nodes
         try:
-            nodes = {persistent[name] if ':' not in name else utils.find_obj(name)
+            nodes = {persistent[name] if ':' not in name else Node.from_dag_node(find_obj(name))
                      for name in selected_nodes
                      if name not in excluded_nodes
                      if name != 'dataset'}
-        except KeyError as e:
-            raise UnknownNode(dataset.collection, e.args[0])
+        except KeyError as exc:
+            raise ConfigError(f'Collection {collection.name!r} has no node {exc}')
 
         if force_dependent:
             nodes.update(x for name in selected_nodes for x in persistent[name].dependent)
@@ -396,7 +267,7 @@ class Site:
             if nodes:
                 changed = await run_nodes(dataset, nodes, store, force=force,
                                           persistent=persistent,
-                                          deps=deps, cachesize=cachesize)
+                                          deps=deps, cachesize=cachesize, site=self)
         finally:
             if not keep:
                 for stream in store.pending:
