@@ -7,18 +7,17 @@ import os
 from itertools import count
 from logging import getLogger
 from pathlib import Path
+from unittest import mock
 
-import mock
 import pytest
 
 import marv.model
 import marv_api as marv
 from marv.app import App
 from marv.site import Site
-from marv.types import Int8Value, Section, Words
+from marv_api.types import Int8Value, Section, Words
 from marv_nodes import dataset as dataset_node
 
-
 log = getLogger(__name__)
 
 
@@ -165,7 +164,7 @@ async def site(loop, request, tmpdir):  # pylint: disable=unused-argument
         class DatetimeBase(vanilla_datetime):
             @classmethod
             def utcnow(cls):
-                ret = vanilla_datetime.fromtimestamp(next(side_effect))
+                ret = vanilla_datetime.utcfromtimestamp(next(side_effect))
                 return ret
 
         return DatetimeMeta('datetime', (DatetimeBase,), {})
@@ -187,12 +186,12 @@ async def site(loop, request, tmpdir):  # pylint: disable=unused-argument
         return Stat()
 
     cls = site_cfg.get('cls', Site)
-    site = await cls.create(str(marv_conf), init=True)  # pylint: disable=redefined-outer-name
+    site = await cls.create(marv_conf, init=True)  # pylint: disable=redefined-outer-name
     try:
         if not site_cfg.get('empty'):
             with mock.patch('bcrypt.gensalt', return_value=b'$2b$12$k67acf6S32i3nW0c7ycwe.'), \
                     mock.patch.object(datetime, 'datetime', create_datemock()), \
-                    mock.patch('marv_node.setid.SetID.random', side_effect=count(2**127)), \
+                    mock.patch('marv_api.setid.SetID.random', side_effect=count(2**127)), \
                     mock.patch('marv.utils.mtime', side_effect=count(1200000000)), \
                     mock.patch('marv.utils.stat', wraps=stat), \
                     mock.patch('marv.utils.now', side_effect=count(1400000000, 12 * 3600)), \
