@@ -12,7 +12,6 @@ from marv.site import Site
 from marv_api import DatasetInfo
 from marv_nodes import dataset as dataset_node
 
-
 DATADIR = Path(__file__).parent / 'data'
 RECORD = os.environ.get('MARV_TESTING_RECORD')
 SETIDS = [
@@ -70,7 +69,7 @@ async def site(loop, tmpdir):  # pylint: disable=unused-argument
             path = tmpdir / 'scanroots' / sitename / name
             path.write(str(idx), ensure=True)
 
-    site_ = await Site.create(marv_conf.strpath, init=True)
+    site_ = await Site.create(marv_conf, init=True)
     yield site_
     await site_.destroy()
 
