@@ -8,22 +8,14 @@ import time
 from datetime import datetime, timedelta
 from datetime import tzinfo as tzinfo_base
 from itertools import islice
-from subprocess import Popen as _Popen
 
-from marv_api.utils import NOTSET, find_obj  # pylint: disable=unused-import
-from marv_node.setid import decode_setid, encode_setid  # pylint: disable=unused-import
+from marv_api import deprecation
+from marv_api.utils import NOTSET
+from marv_api.utils import popen as _popen
 
-
-def echo(*args, **kw):
-    """Wrap print to let linter forbid print usage."""
-    print(*args, **kw)  # noqa: T001
-
-
-def err(*args, exit=None, **kw):
-    """Print to stderr and optionally exit."""
-    print(*args, **kw, file=sys.stderr, flush=True)  # noqa: T001
-    if exit is not None:
-        sys.exit(exit)
+__getattr__ = deprecation.make_getattr(__name__, {
+    'popen': deprecation.Info(__name__, '21.04', _popen, 'use marv_api.utils.popen instead.'),
+})
 
 
 def chunked(iterable, chunk_size):
@@ -127,26 +119,6 @@ def underscore_to_camelCase(string):  # pylint: disable=invalid-name
     return ''.join(x.capitalize() for x in string.split('_'))
 
 
-def sanitize_env(env):
-    ld_library_path = env.get('LD_LIBRARY_PATH')
-    if ld_library_path:
-        clean_path = ':'.join([
-            x
-            for x in ld_library_path.split(':')
-            if not x.startswith('/tmp/_MEI')
-        ])
-        if clean_path:
-            env['LD_LIBRARY_PATH'] = clean_path
-        else:
-            del env['LD_LIBRARY_PATH']
-    return env
-
-
-def popen(*args, env=None, **kw):
-    env = sanitize_env(os.environ.copy() if env is None else env)
-    return _Popen(*args, env=env, **kw)
-
-
 def within_pyinstaller_bundle():
     return any(x for x in sys.path if '/tmp/_MEI' in x)
 
