@@ -3,16 +3,11 @@
 
 from collections import OrderedDict, defaultdict
 from itertools import count
-from logging import getLogger
-
-from capnp.lib.capnp import KjException
 
 from marv_pycapnp import Wrapper
-from .io import Abort
-from .io import CreateStream, Fork, GetRequested, GetStream, SetHeader
-from .io import GetLogger, MakeFile, MsgRequest, Task
-from .io import NEXT, PAUSED, RESUME, THEEND, TheEnd
-from .io import Pull, PullAll, Push
+
+from .io import (NEXT, PAUSED, RESUME, THEEND, CreateStream, Fork, GetRequested, GetStream,
+                 MakeFile, MsgRequest, Pull, PullAll, Push, SetHeader, Task)
 from .mixins import AGenWrapperMixin, LoggerMixin
 from .node import Keyed
 from .stream import Handle, Msg
@@ -49,10 +44,11 @@ class Driver(Keyed, AGenWrapperMixin, LoggerMixin):  # pylint: disable=too-many-
     def name(self):
         return self.stream.name
 
-    def __init__(self, stream, inputs=None):
+    def __init__(self, stream, inputs=None, site=None):
         self.stream = stream
         self.streams = OrderedDict([(stream.handle, stream)])
         self.inputs = inputs
+        self.site = site
         self._requested_streams = []
         self._agen = self._run()
         self._agen_node = None
@@ -69,7 +65,7 @@ class Driver(Keyed, AGenWrapperMixin, LoggerMixin):  # pylint: disable=too-many-
         # pylint: disable=too-many-statements,too-many-branches,too-many-locals
         self.started = True
 
-        agen = self._agen_node = self.node.invoke(self.inputs)
+        agen = self._agen_node = self.node.invoke(self.key_abbrev, self.inputs, site=self.site)
         assert hasattr(agen, 'asend'), agen
 
         yield  # Wait for start signal before returning anything notable
@@ -83,7 +79,7 @@ class Driver(Keyed, AGenWrapperMixin, LoggerMixin):  # pylint: disable=too-many-
         while not finished:
             try:
                 request = await agen.asend(send)
-            except (Abort, StopAsyncIteration):
+            except StopAsyncIteration:
                 finished = True
             else:
                 self.logdebug('got from node %s', type(request))
@@ -123,28 +119,16 @@ class Driver(Keyed, AGenWrapperMixin, LoggerMixin):  # pylint: disable=too-many-
 
                 msg = output
                 if not isinstance(msg, Msg):
-                    msg = self.stream.handle.msg(msg)
+                    msg = self.stream.handle.msg(msg, _schema=self.node.schema)
                 else:
                     assert msg.handle.node is self.node
                     assert msg.handle.setid == self.setid
-                schema = self.node.schema
+
                 # TODO: handles should not be published by all?
                 # this got introduced for merging streams
                 if isinstance(msg._data, Handle):
                     # TODO: check that stream we are publishing to is a Group
                     assert self.node.group, self.node
-                if schema is not None and \
-                   not isinstance(msg._data, (Wrapper, Handle, TheEnd)):
-                    try:
-                        msg._data = Wrapper.from_dict(schema, msg._data)
-                    except KjException as e:
-                        from pprint import pformat  # pylint: disable=import-outside-toplevel
-                        self.logerror(
-                            'Schema violation for %s with data:\n%s\nschema: %s',
-                            schema.schema.node.displayName,
-                            pformat(msg._data),
-                            schema.schema.node.displayName)
-                        raise e
                 signal = yield msg
                 assert signal in (NEXT, RESUME), signal
                 continue
@@ -243,10 +227,6 @@ class Driver(Keyed, AGenWrapperMixin, LoggerMixin):  # pylint: disable=too-many-
                 self.stream_creation = False
                 continue
 
-            if isinstance(request, GetLogger):
-                send = getLogger(f'marv.node.{self.key_abbrev}')
-                continue
-
             raise RuntimeError(
                 f'Unknown request number {request_idx + 1}: {request!r} from {self.node!r}',
             )
@@ -263,8 +243,6 @@ class Driver(Keyed, AGenWrapperMixin, LoggerMixin):  # pylint: disable=too-many-
     async def destroy(self):
         await self._agen_node.aclose()
         await self._agen.aclose()
-        for stream in self.streams.values():
-            stream.destroy()
 
 
 Task.register(Driver)
