@@ -1,15 +1,20 @@
 # Copyright 2016 - 2018  Ternaris.
 # SPDX-License-Identifier: AGPL-3.0-only
 
+import asyncio
 import hashlib
 from base64 import b32encode
 from collections import OrderedDict, namedtuple
+from contextlib import suppress
 from itertools import count, product
+from logging import getLogger
 
 from marv_api import dag
+from marv_api.ioctrl import NODE_SCHEMA, Abort
+from marv_api.iomsgs import GetLogger, GetResourcePath
 from marv_api.utils import find_obj
+
 from . import io
-from .io import get_logger, pull
 from .mixins import Keyed
 
 NODE_CACHE = {}
@@ -139,8 +144,8 @@ class Node(Keyed):  # pylint: disable=too-many-instance-attributes
     def __call__(self, **inputs):
         return self.func(**inputs)
 
-    async def invoke(self, inputs=None):  # noqa: C901
-        # pylint: disable=too-many-locals,too-many-branches
+    async def invoke(self, key_abbrev, inputs=None, site=None):  # noqa: C901
+        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
         # We must not write any instance variables, a node is running
         # multiple times in parallel.
 
@@ -160,7 +165,7 @@ class Node(Keyed):  # pylint: disable=too-many-instance-attributes
                 target.append((spec.name, value))
 
         if foreach_plain or foreach_stream:
-            log = yield get_logger()
+            log = getLogger(f'marv.node.{key_abbrev}')
             cross = product(*[[(k, x) for x in v] for k, v in foreach_plain])
             if foreach_stream:
                 assert len(foreach_stream) == 1, self  # FOR NOW
@@ -168,7 +173,7 @@ class Node(Keyed):  # pylint: disable=too-many-instance-attributes
                 name, stream = foreach_stream[0]
                 idx = count()
                 while True:
-                    value = yield pull(stream)
+                    value = yield io.Pull(stream, False)
                     if value is None:
                         log.noisy('finished forking')
                         break
@@ -192,15 +197,55 @@ class Node(Keyed):  # pylint: disable=too-many-instance-attributes
         else:
             if inputs is None:
                 inputs = dict(common)
-            gen = self.func(**inputs)
-            assert hasattr(gen, 'send')
-            response = None
+            qout, qin = asyncio.Queue(1), asyncio.Queue(1)
+            task = asyncio.create_task(self.execnode(key_abbrev, inputs, qin=qout, qout=qin,
+                                                     site=site),
+                                       name=key_abbrev)
             while True:
+                request = await qin.get()
+                if request is None:
+                    await task
+                    break
                 try:
-                    request = gen.send(response)
-                except StopIteration:
-                    return
-                response = yield request
+                    response = yield request
+                except GeneratorExit:
+                    task.cancel()
+                    with suppress(asyncio.CancelledError):
+                        await task
+                    break
+
+                qout.put_nowait(response)
+
+    async def execnode(self, key_abbrev, inputs, qin, qout, site=None):
+        NODE_SCHEMA.set(self.schema)
+        gen = self.func(**inputs)
+        assert hasattr(gen, 'send')
+        response = None
+        log = getLogger(f'marv.node.{key_abbrev}')
+        while True:
+            try:
+                request = gen.send(response)
+            # TODO: Abort exception needs to invalidate previous node output
+            except (Abort, StopIteration) as exc:
+                msg = str(exc)
+                if msg:
+                    log.warning(msg)
+                qout.put_nowait(None)
+                return
+            except BaseException:  # pylint: disable=broad-except
+                qout.put_nowait(None)
+                raise
+
+            if isinstance(request, GetLogger):
+                response = log
+                continue
+
+            if isinstance(request, GetResourcePath):
+                response = site.config.marv.resourcedir / request.name
+                continue
+
+            qout.put_nowait(request)
+            response = await qin.get()
 
     def __str__(self):
         return self.key
