@@ -6,20 +6,21 @@
 # pylint: disable=too-many-return-statements,too-many-branches,no-else-return,protected-access
 # flake8: noqa
 
+import asyncio
 import functools
 import os
 from collections import OrderedDict, deque
 from logging import getLogger
 from pprint import pformat
 
+from marv_api.setid import SetID
 from marv_node.node import Node
-from marv_node.setid import SetID
 from marv_nodes import dataset as _dataset_node
 from marv_store.streams import ReadStream
+
 from .driver import Driver
 from .event import DefaultOrderedDict
-from .io import MsgRequest, Task
-from .io import NEXT, PAUSED, RESUME, THEEND
+from .io import NEXT, PAUSED, RESUME, THEEND, MsgRequest, Task
 from .stream import Handle, Msg, Stream, VolatileStream
 
 
@@ -27,14 +28,25 @@ class UnmetDependency(Exception):
     pass
 
 
+ABORT = False
 DATASET_NODE = Node.from_dag_node(_dataset_node)
 PPINFO = os.environ.get('PPINFO')
 MARV_RUN_LOGBREAK = os.environ.get('MARV_RUN_LOGBREAK')
 RAISE_IF_UNFINISHED = False
 
 
+def setabort():
+    print('Aborting...')
+    global ABORT  # pylint: disable=global-statement
+    ABORT = True
+
+
+class Aborted(Exception):
+    pass
+
+
 async def run_nodes(dataset, nodes, store, persistent=None, force=None, deps=None, cachesize=None,
-                    _gather_into=None):
+                    site=None, _gather_into=None):
     # pylint: disable=too-many-arguments
 
     if cachesize is not None:
@@ -44,31 +56,39 @@ async def run_nodes(dataset, nodes, store, persistent=None, force=None, deps=Non
 
     queue = []
     ret = await run_nodes_async(dataset=dataset, nodes=nodes, store=store,
-                                queue=queue, persistent=persistent,
+                                queue=queue, persistent=persistent, site=site,
                                 force=force, deps=deps, _gather_into=_gather_into)
     if ret is None:
         return False
 
-    streams, loop, process_task, after = ret
+    streams, loop, process_task, after_loop = ret
 
     # initialize nodes and get initial set of promises
     done = False
     send_queue_empty = False
-    while not done:
-        # comment while for synchronous
-        # while not done and not send_queue_empty:
-        #     # additional random break
-        #     done, send_queue_empty = process_task(None, None)
-        # # random pick from queue
-        current, task = queue.pop(0) if queue else (None, None)
-        done, send_queue_empty = await process_task(current, task)
-    after()
+    try:
+        while not done and not ABORT:
+            # comment while for synchronous
+            # while not done and not send_queue_empty:
+            #     # additional random break
+            #     done, send_queue_empty = process_task(None, None)
+            # # random pick from queue
+            current, task = queue.pop(0) if queue else (None, None)
+            done, send_queue_empty = await process_task(current, task)
+    except BrokenPipeError:  # ffmpeg got killed during abort
+        if not ABORT:
+            raise
+    finally:
+        await after_loop()
+
+    if not done and ABORT:
+        raise Aborted
 
     return streams
 
 
 async def run_nodes_async(dataset, nodes, store, queue, persistent=None, force=None, deps=None,
-                          _gather_into=None):  # noqa: C901
+                          site=None, _gather_into=None):  # noqa: C901
     # pylint: disable=too-many-arguments
     deps = True if deps is None else deps
     force = False if force is None else force
@@ -139,7 +159,7 @@ async def run_nodes_async(dataset, nodes, store, queue, persistent=None, force=N
                 continue
         stream = (store.create_stream(handle) if node in persistent
                   else VolatileStream(handle))
-        driver = Driver(stream)
+        driver = Driver(stream, site=site)
         await start_driver(driver, force and handle in store)
 
     if not drivers:
@@ -407,7 +427,7 @@ async def run_nodes_async(dataset, nodes, store, queue, persistent=None, force=N
                 driver_stream = \
                     store.create_stream(driver_handle) if node in persistent else \
                     VolatileStream(driver_handle)
-                reqdriver = Driver(driver_stream)
+                reqdriver = Driver(driver_stream, site=site)
                 assert reqdriver.key == driver_key
                 await start_driver(reqdriver, forced)
                 if node in persistent:
@@ -420,7 +440,7 @@ async def run_nodes_async(dataset, nodes, store, queue, persistent=None, force=N
 
         return await loop()
 
-    def after_loop():
+    async def after_loop():
         if PPINFO:
             logdebug("state %s", ppinfo())
 
@@ -430,6 +450,8 @@ async def run_nodes_async(dataset, nodes, store, queue, persistent=None, force=N
                       for driver in lst
                       if driver in pulling or driver.node in persistent]
 
+        await asyncio.gather(*[x.destroy() for x in drivers.values()])
+
         if unfinished:
             logdebug("state %s", ppinfo())
             logerror('The following nodes did not finish:\n  %s',
